#![doc = include_str!("../README.md")]
use std::{collections::HashMap, fs, io::{BufWriter, Write}, sync::Mutex};
#[cfg(feature = "parallel")]
use {
    rayon::prelude::{IntoParallelRefIterator, ParallelIterator},
    std::ops::DerefMut,
};

mod checksum;
mod error;
mod file;
mod filetype;
pub(crate) mod nix;
mod node;
mod special;
mod stream;

pub use {
    checksum::{Algorithm, Checksum},
    error::Error,
    file::File,
    filetype::FileType,
    node::Node,
    special::Special,
    stream::Stream,
};

/// Creates a haggis archive from a list of files
pub fn create_archive(path: &str, files: Vec<&str>, algorithm: Algorithm) -> Result<(), Error> {
    let fd = fs::OpenOptions::new()
        .create(true)
        .truncate(true)
        .open(path)?;
    let mut writer = BufWriter::new(fd);
    stream_archive(&mut writer, files, algorithm)?;
    Ok(())
}

/// Streams a haggis archive over something which implements `Write`
pub fn stream_archive<W: Write>(
    mut writer: W,
    files: Vec<&str>,
    algorithm: Algorithm,
) -> Result<(), Error> {
    let links = Mutex::new(HashMap::new());
    for f in &files {
        let node = Node::from_path(f, algorithm, &links)?;
        node.write(&mut writer)?;
    }
    Ok(())
}

/// Creates a Haggis archive from a list of files, processing each file in parallel
#[cfg(feature = "parallel")]
pub fn par_create_archive(path: &str, files: Vec<&str>, algorithm: Algorithm) -> Result<(), Error> {
    let fd = fs::File::create(path)?;
    let writer = BufWriter::new(fd);
    par_stream_archive(writer, files, algorithm)?;
    Ok(())
}

/// Streams a Haggis archive from a list of files, processing each file in parallel
#[cfg(feature = "parallel")]
pub fn par_stream_archive<W: Write + Send>(
    writer: W,
    files: Vec<&str>,
    algorithm: Algorithm,
) -> Result<(), Error> {
    let links = Mutex::new(HashMap::<u64, String>::new());
    let writer = Mutex::new(writer);
    files.par_iter().try_for_each(|f| {
        let node = Node::from_path(f, algorithm, &links)?;
        if let Ok(mut writer) = writer.lock() {
            let mut writer = writer.deref_mut();
            node.write(&mut writer)?;
        }
        Ok::<(), Error>(())
    })?;
    Ok(())
}
